from .retrieval_strategy import RetrievalStrategy
from .graphql_retrieval_strategy import GraphQLRetrievalStrategy
from .influx_retrieval_strategy import InfluxRetrievalStrategy
from .postgres_retrieval_strategy import PostgresRetrievalStrategy



from typing import Optional

import pandas as pd


def average_speed(df: pd.DataFrame, speed_column: str = "SpeedMs") -> Optional[float]:
    """
    Calculate the average speed from a DataFrame.

    Args:
        df (pd.DataFrame): The DataFrame containing the telemetry data.
        speed_column (str): The name of the column containing speed data. Defaults to 'Speed'.

    Returns:
        Optional[float]: The average speed, or None if the speed column is not found or contains no valid data.
    """
    if speed_column not in df.columns:
        print(f"Warning: '{speed_column}' column not found in the DataFrame.")
        return None

    average = df[speed_column].mean()
    if pd.isna(average):
        print(f"Warning: No valid data found in the '{speed_column}' column.")
        return None

    return average


def apex(df: pd.DataFrame, x_column: str = "WorldPosition_x", y_column: str = "WorldPosition_y") -> Optional[dict]:
    """
    Find the apex of a curve in the track.

    Args:
        df (pd.DataFrame): The DataFrame containing the telemetry data.
        x_column (str): The name of the column containing x-coordinate data. Defaults to 'WorldPosition_x'.
        y_column (str): The name of the column containing y-coordinate data. Defaults to 'WorldPosition_y'.

    Returns:
        Optional[dict]: A dictionary containing the index and coordinates of the apex, or None if it can't be found.
    """
    if x_column not in df.columns or y_column not in df.columns:
        print(f"Warning: '{x_column}' or '{y_column}' column not found in the DataFrame.")
        return None

    # remove duplicates
    # df = df.drop_duplicates(subset=[x_column, y_column])

    # Calculate differences
    dx = df[x_column].diff()
    dy = df[y_column].diff()

    # Calculate second differences
    d2x = dx.diff()
    d2y = dy.diff()

    # Calculate curvature
    curvature = (dx * d2y - dy * d2x) / (dx**2 + dy**2) ** (3 / 2)

    # Find the index of the maximum absolute curvature
    apex_index = curvature.abs().idxmax()

    if pd.isna(apex_index):
        print("Warning: Could not determine the apex.")
        return None

    return {"index": apex_index, "x": df.loc[apex_index, x_column], "y": df.loc[apex_index, y_column], "curvature": curvature[apex_index]}



import pandas as pd
import os
from typing import Callable, List, Optional
from plotly.graph_objs import Figure

from racing_telemetry.telemetry import Telemetry
from racing_telemetry.plot import *

def plot_sessions(session_ids: List[str], landmarks: bool = False, columns: Optional[List[str]] = None) -> Figure:
    """
    Plot telemetry data for multiple sessions.

    Args:
        session_ids (list): List of session IDs to plot.
        landmarks (bool, optional): Whether to include landmarks in the plot. Defaults to False.
        columns (list, optional): List of additional columns to include in the plot. Defaults to None.

    Returns:
        plotly.graph_objs._figure.Figure: A Plotly figure object containing the plotted sessions.
    """
    telemetry = Telemetry()
    telemetry.set_pandas_adapter()
    landmark_df = None
    fig = Figure()
    laps = []

    for session_id in session_ids:
        telemetry.set_filter({'session_id': session_id})
        df = get_or_create_df(lambda: telemetry.get_telemetry_df(), name=str(session_id))
        if not df.empty:
            laps.append(df)

    if laps:
        if columns:
            columns.extend(columns)
        fig = lap_fig(laps[0], columns=columns)  # Initialize the figure with the first lap

        for lap in laps[1:]:
            fig = lap_fig(lap, fig=fig, columns=columns)  # Add subsequent laps to the figure

    if landmarks and laps:
        game = laps[0]['GameName'].iloc[0]
        track = laps[0]['TrackCode'].iloc[0]
        landmark_df = telemetry.landmarks(game=game, track=track)
        if landmark_df is not None:
            for _, landmark in landmark_df.iterrows():
                fig = fig_add_shape(fig, x0=landmark['start'], x1=landmark['start'], color='red')

    return fig


def plot_2d_map_sessions(session_ids: List[str], landmarks: bool = False) -> Figure:
    """
    Plot a 2D map of telemetry data for multiple sessions.

    Args:
        session_ids (list): List of session IDs to plot.
        landmarks (bool, optional): Whether to include landmarks in the plot. Defaults to False.

    Returns:
        plotly.graph_objs._figure.Figure: A Plotly figure object containing the 2D map of the sessions.
    """
    telemetry = Telemetry()
    telemetry.set_pandas_adapter()
    landmark_df = None
    fig = Figure()
    laps = []

    for session_id in session_ids:
        telemetry.set_filter({'session_id': session_id})
        df = get_or_create_df(lambda: telemetry.get_telemetry_df(), name=str(session_id))
        if not df.empty:
            laps.append(df)


    if landmarks and laps:
        game = laps[0]['GameName'].iloc[0]
        track = laps[0]['TrackCode'].iloc[0]
        landmark_df = telemetry.landmarks(game=game, track=track)

    if laps:
        fig = plot_2d_map(laps, landmarks=landmark_df)

    return fig


def get_or_create_df(create_df_func: Callable, name: Optional[str] = None) -> pd.DataFrame:
    """
    Get a DataFrame from cache or create a new one if not found in cache.

    Args:
        create_df_func (callable): Function to create the DataFrame if not found in cache.
        name (str, optional): Name to use for the cached file. Defaults to None.

    Returns:
        pandas.DataFrame: The loaded or newly created DataFrame.
    """
    # Get the directory of the calling function

    current_dir: str = os.path.dirname(os.path.abspath(__file__))
    current_dir = os.path.join(current_dir, '../../.cache')
    if name:
        CACHE_FILE: str = os.path.join(current_dir, f'cached_{name}.pkl')
    else:
        CACHE_FILE: str = os.path.join(current_dir, 'cached_df.pkl')

    if os.path.exists(CACHE_FILE):
        print("Loading DataFrame from cache...")
        return pd.read_pickle(CACHE_FILE)

    print("DataFrame not found in cache. Creating a new one...")
    df: pd.DataFrame = create_df_func()

    # Cache the DataFrame
    df.to_pickle(CACHE_FILE)
    print("DataFrame cached to disk.")

    return df








import json
import os
import unittest

import pandas as pd
import vcr

from racing_telemetry.analysis.streaming import Streaming
from racing_telemetry.retrieval.influx_retrieval_strategy import InfluxRetrievalStrategy


class TestStreaming(unittest.TestCase):

    @classmethod
    @vcr.use_cassette("tests/cassettes/test_streaming_session_data.yaml")
    def setUpClass(cls):
        # Use InfluxRetrievalStrategy to get real data
        strategy = InfluxRetrievalStrategy()
        cls.session_data = strategy.retrieve_data(filters=1721368565)

        # Ensure we have data
        assert isinstance(cls.session_data, pd.DataFrame)
        assert len(cls.session_data) > 0

        # Initialize the Streaming class with all features enabled
        cls.streaming = Streaming(average_speed=True, coasting_time=True, raceline_yaw=True, ground_speed=True, braking_point=True, launch_wheel_slip_time=True)

        # Process each row of the session data and collect features
        cls.collected_features = []
        for _, telemetry in cls.session_data.iterrows():
            cls.streaming.notify(telemetry)
            features = cls.streaming.get_features().copy()
            features["CurrentLapTime"] = telemetry["CurrentLapTime"]
            features["Yaw"] = telemetry["Yaw"]
            features["SpeedMs"] = telemetry["SpeedMs"]
            cls.collected_features.append(features)

        # Save or load the collected features
        cls.features_file = "tests/data/test_streaming.json"
        if not os.path.exists(cls.features_file):
            with open(cls.features_file, "w") as f:
                json.dump(cls.collected_features, f, indent=4)
        else:
            with open(cls.features_file, "r") as f:
                cls.expected_features = json.load(f)

    def test_average_speed(self):
        # Compare collected average_speed with expected values
        for i, (collected, expected) in enumerate(zip(self.collected_features, self.expected_features)):
            self.assertIn("average_speed", collected)
            self.assertIsInstance(collected["average_speed"], float)
            self.assertAlmostEqual(collected["average_speed"], expected["average_speed"], places=2, msg=f"Mismatch at index {i}")

    def test_coasting_time(self):
        # Compare collected coasting_time with expected values
        for i, (collected, expected) in enumerate(zip(self.collected_features, self.expected_features)):
            self.assertIn("coasting_time", collected)
            self.assertAlmostEqual(collected["coasting_time"], expected["coasting_time"], places=2, msg=f"Mismatch at index {i}")

        # Calculate total session time
        total_session_time = self.session_data["CurrentLapTime"].max() - self.session_data["CurrentLapTime"].min()

        # Coasting time should be less than or equal to total session time
        self.assertLessEqual(self.collected_features[-1]["coasting_time"], total_session_time)

    def test_raceline_yaw(self):
        # Compare collected raceline_yaw with expected values
        for i, (collected, expected) in enumerate(zip(self.collected_features, self.expected_features)):
            self.assertIn("raceline_yaw", collected)
            self.assertIsInstance(collected["raceline_yaw"], float)
            self.assertAlmostEqual(collected["raceline_yaw"], expected["raceline_yaw"], places=2, msg=f"Mismatch at index {i}")
            # Check if the yaw is within the expected range (-180 to 180 degrees)
            self.assertGreaterEqual(collected["raceline_yaw"], -180)
            self.assertLessEqual(collected["raceline_yaw"], 180)

    def test_ground_speed(self):
        # Compare collected ground_speed with expected values
        for i, (collected, expected) in enumerate(zip(self.collected_features, self.expected_features)):
            self.assertIn("ground_speed", collected)
            self.assertIsInstance(collected["ground_speed"], float)
            self.assertAlmostEqual(collected["ground_speed"], expected["ground_speed"], places=2, msg=f"Mismatch at index {i}")
            # Check if the ground speed is non-negative
            self.assertGreaterEqual(collected["ground_speed"], 0)

    def test_braking_point(self):
        # Find the first occurrence of a valid braking point
        first_valid_braking_point = next((feature["braking_point"] for feature in self.collected_features if feature["braking_point"] != -1), None)

        # Ensure a valid braking point was found
        self.assertIsNotNone(first_valid_braking_point, "No valid braking point was detected")

        # Check that the braking point is within the expected range
        self.assertGreater(first_valid_braking_point, 0)
        self.assertLess(first_valid_braking_point, self.session_data["DistanceRoundTrack"].max())

        # Verify that the braking point remains constant after it's first detected
        braking_points = [feature["braking_point"] for feature in self.collected_features]
        first_valid_index = braking_points.index(first_valid_braking_point)
        self.assertTrue(all(bp == first_valid_braking_point for bp in braking_points[first_valid_index:]), "Braking point should remain constant after first detection")

        # Verify that all values before the first valid braking point are -1
        self.assertTrue(all(bp == -1 for bp in braking_points[:first_valid_index]), "All braking point values before detection should be -1")

    def test_launch_wheel_slip_duration(self):
        # Compare collected launch_wheel_slip_duration with expected values
        for i, (collected, expected) in enumerate(zip(self.collected_features, self.expected_features)):
            self.assertIn("launch_wheel_slip_duration", collected)
            self.assertIsInstance(collected["launch_wheel_slip_duration"], float)
            self.assertAlmostEqual(collected["launch_wheel_slip_duration"], expected["launch_wheel_slip_duration"], places=2, msg=f"Mismatch at index {i}")

        # Check that the launch_wheel_slip_duration is non-negative
        self.assertGreaterEqual(self.collected_features[-1]["launch_wheel_slip_duration"], 0)

        # Check that the launch_wheel_slip_duration is monotonically increasing
        durations = [feature["launch_wheel_slip_duration"] for feature in self.collected_features]
        self.assertEqual(durations, sorted(durations), "launch_wheel_slip_duration should be monotonically increasing")


if __name__ == "__main__":
    unittest.main()



import math
from typing import Callable, Dict

import pandas as pd
from loguru import logger  # noqa: F401

import racing_telemetry.analysis.basic_stats as basic_stats


class Streaming:
    def __init__(
        self,
        average_speed: bool = False,
        coasting_time: bool = False,
        raceline_yaw: bool = False,
        ground_speed: bool = False,
        braking_point: bool = False,
        wheel_slip: bool = False,
        lift_off_point: bool = False,
        acceleration_point: bool = False,
        launch_wheel_slip_time: bool = False,
        **kwargs,
    ):
        self.features: Dict[str, Callable] = {}
        self.computed_features: Dict[str, float] = {}

        if average_speed:
            self.configure_feature("average_speed", self.average_speed)
        if coasting_time:
            self.configure_feature("coasting_time", self.coasting_time)
        if raceline_yaw:
            self.configure_feature("raceline_yaw", self.raceline_yaw)
        if braking_point:
            self.configure_feature("braking_point", self.braking_point)
        if launch_wheel_slip_time:
            # wheel_slip requires ground_speed to be computed
            self.configure_feature("ground_speed", self.ground_speed)
            self.configure_feature("wheel_slip", self.wheel_slip)
            self.configure_feature("launch_wheel_slip_time", self.launch_wheel_slip_time)
            wheel_slip = False
            ground_speed = False
        if wheel_slip:
            # wheel_slip requires ground_speed to be computed
            self.configure_feature("ground_speed", self.ground_speed)
            self.configure_feature("wheel_slip", self.wheel_slip)
            ground_speed = False
        if ground_speed:
            self.configure_feature("ground_speed", self.ground_speed)
        if lift_off_point:
            self.configure_feature("lift_off_point", self.lift_off_point)
        if acceleration_point:
            self.configure_feature("acceleration_point", self.acceleration_point)

        self.telemetry: Dict[str, list] = {
            "CurrentLapTime": [],
            "WorldPosition_x": [],
            "WorldPosition_y": [],
            "Throttle": [],
            "Brake": [],
            "SpeedMs": [],
            "DistanceRoundTrack": [],
        }
        self.telemetry_len = 0
        self.brake_pressed: bool = False
        self.braking_point_found: bool = False
        self.lift_off_point_found: bool = False
        self.total_coasting_time: float = 0.0
        self.brake_pressure_threshold: float = 0.1
        self.rate_of_change_threshold: float = 0.05
        self.throttle_threshold: float = 0.9
        self.throttle_decrease_threshold: float = 0.1
        self.acceleration_point_found: bool = False
        self.throttle_increase_threshold: float = 0.1
        self.speed_increase_threshold: float = 0.5
        self.current_wheel_slip_time: float = 0.0

    def configure_feature(self, name: str, feature_func: Callable):
        """
        Configure a new feature to be computed.

        Args:
            name (str): The name of the feature.
            feature_func (Callable): The function to compute the feature.
        """
        self.features[name] = feature_func
        self.computed_features[name] = 0.0

    def notify(self, telemetry: Dict):
        """
        Process new telemetry data and compute configured features.

        Args:
            telemetry (Dict): The incoming telemetry data.
        """
        self._store_telemetry_values(telemetry)
        self.delta_time = self._calculate_elapsed_time()
        self.delta_x, self.delta_y = self._calculate_position_delta(telemetry.get("WorldPosition_x", 0.0), telemetry.get("WorldPosition_y", 0.0))

        self._calculate_change_rate()

        for feature_name, feature_func in self.features.items():
            result = feature_func(telemetry)
            if result is not None:
                self.computed_features[feature_name] = result

    def get_features(self) -> Dict[str, float]:
        """
        Get the computed features.

        Returns:
            Dict[str, List[float]]: A dictionary of feature names and their computed values.
        """
        return self.computed_features

    def _calculate_elapsed_time(self) -> float:
        """Calculate elapsed time since last update."""
        if self.telemetry_len < 2:
            return 0

        current_lap_time = self.telemetry["CurrentLapTime"][-1]
        last_lap_time = self.telemetry["CurrentLapTime"][-2]
        elapsed_time = current_lap_time - last_lap_time
        return elapsed_time

    def _calculate_position_delta(self, current_x: float, current_y: float) -> tuple[float, float]:
        """Calculate the change in position since last update."""
        if len(self.telemetry["WorldPosition_x"]) < 2:
            return 0.0, 0.0
        last_x = self.telemetry["WorldPosition_x"][-2]
        last_y = self.telemetry["WorldPosition_y"][-2]
        dx = current_x - last_x
        dy = current_y - last_y
        return dx, dy

    def _calculate_change_rate(self) -> None:
        """Calculate the change rate of throttle and brake values over the last second."""
        if self.telemetry_len < 2:
            self.throttle_change_rate = 0.0
            self.brake_change_rate = 0.0
        else:
            current_throttle = self.telemetry["Throttle"][-1]
            current_brake = self.telemetry["Brake"][-1]
            last_throttle = self.telemetry["Throttle"][-2]
            last_brake = self.telemetry["Brake"][-2]
            self.throttle_change_rate = current_throttle - last_throttle
            self.brake_change_rate = current_brake - last_brake

    def _store_telemetry_values(self, telemetry: Dict) -> None:
        """Store all values of the telemetry dict in a dict of arrays."""
        for key in self.telemetry.keys():
            value = telemetry.get(key, None)
            self.telemetry[key].append(value)
        self.telemetry_len += 1

    def average_speed(self, telemetry: Dict) -> float:
        """
        Calculate the running average speed.

        Args:
            current_speed (float): The current speed value.

        Returns:
            Optional[float]: The updated average speed, or None if no data has been processed.
        """

        return sum(self.telemetry["SpeedMs"]) / self.telemetry_len

    def coasting_time(self, telemetry: Dict) -> float:
        """
        Calculate the time spent coasting (no Throttle or Brake applied).

        Args:
            telemetry (Dict): The incoming telemetry data.

        Returns:
            float: The total time spent coasting in seconds.
        """
        if telemetry.get("Throttle", 0) == 0 and telemetry.get("Brake", 0) == 0:
            self.total_coasting_time += self.delta_time
        return self.total_coasting_time

    def raceline_yaw(self, telemetry: Dict) -> float:
        """
        Calculate the yaw based on the current and previous x and y coordinates.

        Args:
            telemetry (Dict): The incoming telemetry data.

        Returns:
            float: The calculated yaw angle between -180 and 180 degrees.
        """

        dx, dy = self.delta_x, self.delta_y

        if dx == 0 and dy == 0:
            return 0.0

        yaw = math.degrees(math.atan2(dy, dx))

        yaw = (yaw - 90) % 360
        if yaw > 180:
            yaw -= 360

        return yaw

    def ground_speed(self, telemetry: Dict) -> float:
        """
        Calculate the ground speed based on x and y coordinates traveled between ticks.

        Args:
            telemetry (Dict): The incoming telemetry data.

        Returns:
            float: The calculated ground speed in meters per second.
        """

        if self.delta_time == 0:
            return 0.0

        dx, dy = self.delta_x, self.delta_y

        distance = math.sqrt(dx**2 + dy**2)
        speed = distance / self.delta_time

        return speed

    def braking_point(self, telemetry: Dict) -> float:
        """
        Determine the braking point based on brake pressure and its rate of change.

        Args:
            telemetry (Dict): The incoming telemetry data.

        Returns:
            float: The DistanceRoundTrack when the braking point is detected, or -1 if not yet detected.
        """
        if self.braking_point_found:
            return self.computed_features["braking_point"]

        if self.telemetry_len < 2:
            return -1

        if self.delta_time == 0:
            return -1

        current_brake_pressure = telemetry.get("Brake", 0)

        if current_brake_pressure > self.brake_pressure_threshold and self.brake_change_rate > self.rate_of_change_threshold:
            self.braking_point_found = True
            return telemetry.get("DistanceRoundTrack", -1)
        return -1

    def wheel_slip(self, telemetry: Dict) -> float:
        """
        Calculate the wheel slip based on the difference between ground speed and SpeedMs.

        Args:
            telemetry (Dict): The incoming telemetry data.

        Returns:
            float: The wheel slip as a percentage between -1 and 1.
        """
        ground_speed = self.computed_features.get("ground_speed", 0.0)
        speed_ms = telemetry.get("SpeedMs", 0)

        if speed_ms == 0:
            return 0.0

        slip = (ground_speed - speed_ms) / speed_ms
        return max(min(slip, 1.0), -1.0)

    def lift_off_point(self, telemetry: Dict) -> float:
        """
        Determine the lift-off point based on throttle decrease.

        Args:
            telemetry (Dict): The incoming telemetry data.

        Returns:
            float: The DistanceRoundTrack when the lift-off point is detected, or -1 if not yet detected.
        """
        if self.lift_off_point_found:
            return self.computed_features["lift_off_point"]

        if self.telemetry_len < 2:
            return -1

        current_throttle = telemetry.get("Throttle", 0)
        last_throttle = self.telemetry["Throttle"][-2]

        if last_throttle > self.throttle_threshold and (last_throttle - current_throttle) > self.throttle_decrease_threshold:
            self.lift_off_point_found = True
            return telemetry.get("DistanceRoundTrack", -1)
        return -1

    def acceleration_point(self, telemetry: Dict) -> float:
        """
        Determine the acceleration point based on throttle increase and speed increase.

        Args:
            telemetry (Dict): The incoming telemetry data.

        Returns:
            float: The DistanceRoundTrack when the acceleration point is detected, or -1 if not yet detected.
        """
        if self.acceleration_point_found:
            return self.computed_features["acceleration_point"]

        if self.telemetry_len < 2:
            return -1

        current_throttle = telemetry.get("Throttle", 0)
        current_speed = telemetry.get("SpeedMs", 0)
        last_throttle = self.telemetry["Throttle"][-2]
        last_speed = self.telemetry["SpeedMs"][-2]

        throttle_increase = current_throttle - last_throttle
        speed_increase = current_speed - last_speed

        if throttle_increase > self.throttle_increase_threshold and speed_increase > self.speed_increase_threshold:
            self.acceleration_point_found = True
            return telemetry.get("DistanceRoundTrack", -1)

        return -1

    def launch_wheel_slip_time(self, telemetry: Dict) -> float:
        """
        Calculate the duration of wheel slip.

        Args:
            telemetry (Dict): The incoming telemetry data.

        Returns:
            float: The duration of wheel slip in seconds, or 0 if below the threshold.
        """

        if self.telemetry_len > 2:
            current_lap_time = telemetry.get("CurrentLapTime", 0.0)
            # Only consider the first 5 seconds of the lap
            if current_lap_time < 5:
                current_wheel_slip = self.computed_features.get("wheel_slip", 0.0)
                if abs(current_wheel_slip) > 0.1:
                    self.current_wheel_slip_time += self.delta_time
        return self.current_wheel_slip_time

    def calculate_apex(self) -> None:
        """
        Trigger the apex calculation.

        Returns:
            Dict: The apex information or None if it can't be calculated.
        """
        x_positions = self.telemetry.get("WorldPosition_x", [])
        y_positions = self.telemetry.get("WorldPosition_y", [])
        df = pd.DataFrame({"WorldPosition_x": x_positions, "WorldPosition_y": y_positions})
        apex = basic_stats.apex(df)
        apex_distance = -1
        if apex:
            distance_round_track = self.telemetry.get("DistanceRoundTrack", [])
            apex_distance = distance_round_track[apex["index"]]
        self.computed_features["apex"] = apex_distance



from .retrieval import *
from .adapter import *

from typing import Optional, Any, Dict

class Telemetry:
    def __init__(self) -> None:
        self.filter: Dict[str, Any] = {}
        self.graphql_strategy: GraphQLRetrievalStrategy = GraphQLRetrievalStrategy()
        self.influx_strategy: InfluxRetrievalStrategy = InfluxRetrievalStrategy()
        self.postgres_strategy: PostgresRetrievalStrategy = PostgresRetrievalStrategy()
        self.adapter: Adapter = TransparentAdapter()

    def set_pandas_adapter(self) -> None:
        self.adapter = PandasAdapter()

    def games(self) -> Any:
        return self.adapter.convert(
            self.postgres_strategy.games()
        )

    def sessions(self, group_by: Optional[str] = None, limit: Optional[int] = 10,
                 game: Optional[str] = None, track: Optional[str] = None, driver: Optional[str] = None) -> Any:
        return self.adapter.convert(
            self.postgres_strategy.sessions(group_by=group_by, limit=limit, game_name=game, track_name=track, driver_name=driver)
        )

    def drivers(self) -> Any:
        return self.adapter.convert(
            self.postgres_strategy.drivers()
        )

    def tracks(self, game: Optional[str] = None, track: Optional[str] = None) -> Any:
        return self.adapter.convert(
            self.postgres_strategy.tracks(game_name=game, track_name=track)
        )

    def landmarks(self, game: Optional[str] = None, track: Optional[str] = None, kind: Optional[str] = None) -> Any:
        return self.adapter.convert(
            self.postgres_strategy.landmarks(game_name=game, track_name=track, kind=kind)
        )

    def cars(self, game: Optional[str] = None) -> Any:
        return self.adapter.convert(
            self.postgres_strategy.cars(game=game)
        )

    def set_filter(self, filter: Dict[str, Any]) -> None:
        self.filter = filter

    def get_data(self, adapter: Adapter = TransparentAdapter()) -> Any:
        raw_data = self.graphql_strategy.retrieve_data(self.filter)
        return adapter.convert(raw_data)

    def get_data_df(self) -> Any:
        return self.get_data(adapter=PandasAdapter())

    def get_telemetry(self, adapter: Adapter = TransparentAdapter()) -> Any:
        self.strategy: InfluxRetrievalStrategy = InfluxRetrievalStrategy()
        raw_data = self.strategy.retrieve_data(self.filter)
        return adapter.convert(raw_data)

    def get_telemetry_df(self) -> Any:
        return self.get_telemetry(adapter=PandasAdapter())



from datetime import datetime
import os
import pandas as pd

import influxdb_client
from .retrieval_strategy import RetrievalStrategy

class InfluxRetrievalStrategy(RetrievalStrategy):

    def __init__(self) -> None:
        (self.org, self.token, self.url) = self._get_influxdb2_config()

        self.influx = influxdb_client.InfluxDBClient(
            url=self.url, token=self.token, org=self.org, timeout=(10_000, 600_000)
        )
        # if self.influx.ping():
        #     logging.debug(f"Influx: Connected to {self.url}")
        # else:
        #     logging.error(f"Influx: Connection to {self.url} failed")
        #     sys.exit(1)

        self.query_api = self.influx.query_api()


    def _get_influxdb2_config(self) -> tuple[str, str, str] | Exception:
        """Get InfluxDB2 configuration from environment variables."""
        org = os.environ.get("B4MAD_RACING_INFLUX_ORG", "b4mad")

        _INFLUXDB2_TOKEN = os.environ.get("B4MAD_RACING_INFLUX_TOKEN")  # noqa: N806

        if _INFLUXDB2_TOKEN is None or _INFLUXDB2_TOKEN == "":
            raise Exception(
                "B4MAD_RACING_INFLUX_TOKEN",
            )

        token = _INFLUXDB2_TOKEN

        _INFLUXDB2_SERVICE_HOST = os.environ.get("INFLUXDB2_SERVICE_HOST")  # noqa: N806
        _INFLUXDB2_SERVICE_PORT = os.environ.get("INFLUXDB2_SERVICE_PORT", 8086)  # noqa: N806
        _INFLUXDB2_SERVICE_PROTOCOL = os.environ.get("INFLUXDB2_SERVICE_PROTOCOL", "http")  # noqa: N806

        if _INFLUXDB2_SERVICE_HOST is None:
            raise Exception(
                "INFLUXDB2_SERVICE_HOST",
            )

        url = f"{_INFLUXDB2_SERVICE_PROTOCOL}://{_INFLUXDB2_SERVICE_HOST}:{_INFLUXDB2_SERVICE_PORT}/"

        return (org, token, url)

    def retrieve_data(self, filters):
        # if filters is an integer
        if isinstance(filters, int):
            return self.session_df(session_id=filters)
        if isinstance(filters, dict):
            return self.session_df(**filters)
            # if "session_id" in filters:
            #     return self.session_df(session_id=filters["session_id"])
            # if "session_id" in filters and "lap_number" in filters:
            #     return self.session_df(
            #         session_id=filters["session_id"], lap_number=filters["lap_number"]
            #     )
        return None

    def session_df(self, **kwargs):
        kwargs["measurement"] = "fast_laps"
        kwargs["bucket"] = "fast_laps"
        df = self._session_df(**kwargs)
        if df.empty:
            kwargs["measurement"] = "laps_cc"
            kwargs["bucket"] = "racing"
            df = self._session_df(**kwargs)
        return df

    def _session_df(
        self,
        session_id,
        lap_number=None,
        start="-365d",
        end="now()",
        measurement="fast_laps",
        bucket="fast_laps",
        aggregate="",
        fields=[],
        drop_tags=False,
        driver="",
        session_type="Race"
    ):
        if isinstance(start, datetime):
            start = start.strftime("%Y-%m-%dT%H:%M:%S.%fZ")
        if isinstance(end, datetime):
            end = end.strftime("%Y-%m-%dT%H:%M:%S.%fZ")

        query = f"""
        from(bucket: "{bucket}")
        |> range(start: {start}, stop: {end})
        |> filter(fn: (r) => r["_measurement"] == "{measurement}")
        |> filter(fn: (r) => r["SessionId"] == "{session_id}")
        """

        if fields:
            query += f"""
            |> filter(fn: (r) => r["_field"] == "{fields[0]}")
            """
            for field in fields[1:]:
                query += f"""
                or r["_field"] == "{field}"
                """
            query += ")\n"

        if session_type:
            query += f"""
            |> filter(fn: (r) => r["SessionTypeName"] == "{session_type}")
            """

        if driver:
            query += f"""
            |> filter(fn: (r) => r["user"] == "{driver}")
            """

        if aggregate:
            # downsample to 1Hz
            query += f"""
            |> aggregateWindow(every: {aggregate}, fn: last, createEmpty: false)
            """

        query += """
        |> pivot(rowKey: ["_time"], columnKey: ["_field"], valueColumn: "_value")
        """

        if lap_number:
            query += f"""
        |> filter(fn: (r) => r["CurrentLap"] == "{lap_number}")
        """

        query += """
        |> sort(columns: ["_time"], desc: false)
        """

        # drop columns that are not needed
        if drop_tags:
            query += """
            |> drop(columns: ["_start", "_stop", "_measurement", "host", "topic", "user"])
            """

        # print(query)
        df = self.query_api.query_data_frame(query=query)
        # set pd options to display all columns
        # pd.set_option("display.max_columns", None)
        # print(df)
        if df.empty:
            return df

        game = df["GameName"].iloc[0]
        if game == "Assetto Corsa Competizione":
            # flip y axis
            df["x"] = df["WorldPosition_x"]
            df["y"] = df["WorldPosition_z"] * -1
            df["WorldPosition_x"] = df["x"]
            df["WorldPosition_y"] = df["y"]
        if game == "Automobilista 2":
            df["x"] = df["WorldPosition_x"]
            df["y"] = df["WorldPosition_z"]
            df["WorldPosition_x"] = df["x"]
            df["WorldPosition_y"] = df["y"]
        if game == "Richard Burns Rally":
            # rotate 90 degrees to the left
            df["x"] = df["WorldPosition_y"]
            df["y"] = df["WorldPosition_x"]
            df["WorldPosition_x"] = -df["x"]
            df["WorldPosition_y"] = df["y"]


        df["id"] = df["SessionId"].astype(str) + "-" + df["CurrentLap"].astype(str)

        df = df[df["Gear"] != 0]

        return df



import pandas as pd
from .adapter import Adapter

class PandasAdapter(Adapter):
    def convert(self, data):
        # if data is a DataFrame, just return it
        if isinstance(data, pd.DataFrame):
            return data
        # Implement conversion to pandas DataFrame here
        converted_data = []
        for entry in data:
            if hasattr(entry, '__table__'):
                entry_dict = {column.name: getattr(entry, column.name) for column in entry.__table__.columns}
                converted_data.append(entry_dict)
            else:
                converted_data.append(entry)
        return pd.DataFrame(converted_data)



from setuptools import find_packages, setup

setup(
    name="racing-telemetry",
    version="0.1.6",
    packages=find_packages(include=["racing_telemetry", "racing_telemetry.*"]),
    install_requires=[
        "gql",
        "requests",
        "influxdb-client",
        "pandas",
        "requests-toolbelt",
        "psycopg2-binary",
        "sqlalchemy",
    ],
    author="Marcel Hild",
    author_email="hild@b4mad.net",
    description="A library for telemetry data analysis",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/b4mad/racing-telemetry",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
)
